# Copyright (C) 2009 Daniel Maturana
# This file is part of subspaceextract.
#
# subspaceextract is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# subspaceextract is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with subspaceextract. If not, see <http://www.gnu.org/licenses/>.
#

import numpy as np
import scipy.io
import scipy.linalg as la
import random
from dataset import Dataset

# for some reason it's not in numpy? it's in scipy.linalg.basic
eps = 2.22044604925e-16

def null(A, tol=None):
    """ Nullspace of matrix A """
    (U, s, Vh) = la.svd(A)
    # note: numpy's s is the diagonal of matlab's S
    rows, cols = A.shape
    if tol is None:
        tol = np.max(np.array(A.shape)*s[0]*eps)
    rank = (s > tol).sum()
    if rank >= cols:
        return np.zeros((cols, 0))
    ns = Vh[rank:,:].T
    # zero out small vectors
    ns[np.abs(ns) < eps] = 0
    return ns

def largest_eig(A, B, num):
    """ Find largest eigenvector of A or generalized vectors A, B """
    if B is None:
        (eigvals, eigvecs) = la.eig(A)
    else:
        (eigvals, eigvecs) = la.eig(A, B)
    # sort by eigval size, decreasing, keep largest
    ix = np.argsort(-eigvals)
    eigvals = eigvals[ix][:num]
    eigvecs = eigvecs[:,ix][:,:num]
    # unitize columns
    eigvecs /= np.sqrt((eigvecs**2).sum(0))
    return eigvals, eigvecs

def subspace_project(W, mean, data):
    """ Returns coefficients of data when projected onto space.
    This is the 'compressed' version of the original data.
    In: data[NxD], mean[D], W[DxD']
    Out: pdata[NxD']
    """
    # if data is a single vector treat is a single row 2d array
    data = np.atleast_2d(data)
    pdata = np.dot(data - mean, W)
    return pdata

def subspace_reconstruct(W, mean, pdata):
    """ Reconstruct data from subspace (as best as possible).
    In: W[DxD'], pdata[NxD'], mean[D]
    Out: rdata[NxD]
    """
    # if data is a single vector treat is a single row 2d array
    pdata = np.atleast_2d(pdata)
    rdata = np.dot(pdata, W.T) + mean
    return rdata

def scatter_matrices(dataset):
    """ Intra-class and inter-class scatter matrices. """
    # calculate class stats (mean, sizes)
    centered_data = dataset.data - dataset.mean
    class_sizes = dict([(c, 0) for c in dataset.classes])
    class_means = dict([(c, 0.0) for c in dataset.classes])
    # This can probably be done more efficiently
    for (x, label) in zip(centered_data, dataset.labels):
        class_sizes[label] += 1
        class_means[label] += x
    for c in class_means.keys():
        class_means[c] /= class_sizes[c]
    # Compute between class scatter matrix, Sb
    print dataset.D
    Sb = np.zeros((dataset.D,dataset.D))
    for c in class_means.keys():
        mc = class_means[c]
        Sb += (class_sizes[c]*np.outer(mc,mc))
    # Compute within class scatter matrix, Sw
    Sw = np.zeros((dataset.D,dataset.D))
    # This is the same as calculating Sc separately for each class
    # (Sc) and add them: Sw = sum(Sc, c=1..C).
    for n,x in enumerate(centered_data):
        d = x-class_means[dataset.labels[n]]
        Sw += np.outer(d,d)
    return (Sb, Sw)

def pca(dataset, D_pca):
    """ Find principal components of data.
    Wpca is the embedding matrix for pca, ie
    if X is the data matrix Xlowdim = Wpca*X
    """
    centered_data = dataset.data - dataset.mean
    if dataset.D > dataset.N:
        # A*A.T is too large -- instead use A.T*A for cov
        cov = np.dot(centered_data, centered_data.T)
        eigvals, eigvecs = largest_eig(cov, None, D_pca)
        del cov
        # now get eigvecs of A.T*A
        eigvecs = np.dot(centered_data.T, eigvecs)
    else:
        # more typical case, use usual covar matrix
        cov = np.dot(centered_data.T, centered_data)
        eigvals, eigvecs = largest_eig(cov, None, D_pca)
        del cov
    # unitize columns
    eigvecs /= np.sqrt((eigvecs**2).sum(0))
    return eigvecs

def lda(dataset, D_lda):
    """ Find W to maximize J(W)=det((W.T*Sb*W)/(W.T*Sw*W)).
    """
    if D_lda > dataset.C-1:
        raise ValueError("Max number of Lda features is C-1")
    (Sb, Sw) = scatter_matrices(dataset)
    # Solve the generalized eigenvector equation
    # Sb*w_i = lambda_i*Sw*w_i, where w_i are cols of W
    (eigvals, Wlda) = largest_eig(Sb, Sw, D_lda)
    # unitize columns
    Wlda /= np.sqrt((Wlda**2).sum(0))
    return Wlda

def fisher_faces(dataset, D_lda):
    """ Perform LDA on data using fisherface technique.
    """
    # Use the PCA matrix on the data and then calculate the scatter matrices
    # on the reduced dimensionality data
    D_pca = dataset.N-dataset.C
    Wpca = pca(dataset, D_pca)
    pdata = subspace_project(Wpca, dataset.mean, dataset.data)
    pdataset = Dataset(pdata, dataset.labels)
    Wlda = lda(pdataset, D_lda)
    # Project back onto original space
    W = np.dot(Wpca, Wlda)
    # Unitize columns
    W /= np.sqrt((W**2).sum(0))
    return W

def random_lda(dataset, D_lda):
    """ Find an LDA matrix via random sampling LDA technique.
    Uses Wang and Tang 2006 random sampling LDA - first
    find eigenfaces and then find LDA on a random subset of
    eigenfaces.
    """
    if D_lda is None:
        D_lda = dataset.C-1
    elif D_lda > dataset.C-1:
        # max dimensionality of eigenspace is C-1
        raise ValueError("Max number of Lda features is C-1")
    # Extract PCA matrix with N-1 eigenvectors (rank of covar matrix)
    Wpca = pca(dataset, dataset.N-1)
    # Intermediate dimensionality reduction to D_pca.
    # W&T don't specify how they set D_pca - it should be less than N-C.
    # We use a rule of thumb.
    D_intermediate = min(4*(dataset.N-dataset.C)/5,100)
    # As a base we use D0 largest vectors from Wpca and D1 random
    # eigenvectors from Wpca which aren't among the D0 largest.
    # W&T use D0 = D1 = D_intermediate/2.
    D0 = D_intermediate/2
    D1 = D_intermediate - D0
    # largest D0 eigenvector indices
    d0_col_ix = range(D0)
    # random D1 eigenvector indices
    rand_col_ix = random.sample(xrange(D0, Wpca.shape[1]), D1)
    # combined indices
    col_ix = d0_col_ix + rand_col_ix
    Wrandom = Wpca[:,col_ix]
    # Project data onto this semi-random subspace and do LDA
    pdata = subspace_project(Wrandom, dataset.mean, dataset.data)
    pdataset = Dataset(pdata, dataset.labels)
    Wlda = lda(pdataset, D_lda)
    # Project fisher vectors onto original space
    W = np.dot(Wrandom, Wlda)
    # unitize columns of W
    W /= np.sqrt((W**2).sum(0))
    # now W's columns are 'randomldafaces'
    return W

def null_lda(dataset, D_lda):
    """ Nullspace LDA. """
    # max dimensionality of eigenspace is C-1
    if D_lda > dataset.C-1:
        raise ValueError("Max number of NLda features is C-1")
    if D_lda is None:
        D_lda = dataset.C-1
    # Use the PCA matrix on the data and then calculate the scatter matrices
    # on the reduced dimensionality data
    D_pca = dataset.N-1
    Wpca = pca(dataset, D_pca)
    pdata = subspace_project(Wpca, dataset.mean, dataset.data)
    pdataset = Dataset(pdata, dataset.labels)
    (Sb, Sw) = scatter_matrices(pdataset)
    # Find nullspace of Sw
    nullSw = null(Sw)
    Sb2 = np.dot(np.dot(nullSw.T, Sb), nullSw)
    (eigvals, phi) = largest_eig(Sb2, None, dataset.C-1)
    Wnld = np.dot(nullSw, phi)
    W = np.dot(Wpca, Wnld)
    # unitize columns of W
    W /= np.sqrt((W**2).sum(0))
    # now W's columns are 'fisherfaces'
    return W

def bagged_null_lda(dataset, out_D):
    """ Bagged nullspace LDA. """
    # max dimensionality of eigenspace is C-1
    if out_D > dataset.C-1:
        raise ValueError("Max number of Bagged NLda features is C-1")
    if out_D is None:
        out_D = dataset.C-1
    #L1 = dataset.C/2
    sclasses = random.sample(dataset.classes, out_D+1)
    ix = np.zeros(dataset.N, dtype=np.bool)
    # select ix s.t. labels == class for each class
    for c in sclasses:
        ix = ix | (dataset.labels==c)
    bagged_dataset = Dataset(dataset.data[ix], dataset.labels[ix])
    #pdb.set_trace()
    Wnlda = null_lda(bagged_dataset, out_D)
    return Wnlda

class BaseSubspace(object):
    """ Base subspace object. """
    def __init__(self, out_D=None):
        self.out_D = out_D

    def extract(self, dataset):
        """ Extract subspace basis from dataset. """
        raise NotImplementedError();

    def project(self, dataset):
        """ Project dataset on subspace basis. """
        pdata = subspace_project(self.basis, self.mean, dataset.data)
        return Dataset(pdata, dataset.labels)

    def reconstruct(self, pdataset):
        """ Reconstruct dataset from basis vectors. """
        rdata = subspace_reconstruct(self.basis, self.mean, pdataset.data)
        return Dataset(rdata, pdataset.labels)

    def save(self, filename, format='npz'):
        """ Save basis in file. """
        savedict = {'basis':self.basis, 'mean':self.mean}
        if format == 'npz':
            np.savez(filename, **savedict)
        elif format == 'mat':
            scipy.io.savemat(filename, savedict)
        else:
            raise ValueError("%s format not recognized"%format)

class FisherFaceSubspace(BaseSubspace):
    def extract(self, dataset):
        self.basis = fisher_faces(dataset, self.out_D)
        self.mean = dataset.mean

class LdaSubspace(BaseSubspace):
    def extract(self, dataset):
        self.basis = lda(dataset, self.out_D)
        self.mean = dataset.mean

class PcaSubspace(BaseSubspace):
    def extract(self, dataset):
        self.basis = pca(dataset, self.out_D)
        self.mean = dataset.mean

class RandomLdaSubspace(BaseSubspace):
    def extract(self, dataset):
        self.basis = random_lda(dataset, self.out_D)
        self.mean = dataset.mean

class BaggedNullLdaSubspace(BaseSubspace):
    def extract(self, dataset):
        self.basis = bagged_null_lda(dataset, self.out_D)
        self.mean = dataset.mean
